<?php

use ctx\traits\LibraryLoader;
use ctx\traits\StoreLoader;

class ControllerExtensionConnectixApi extends Controller
{
    use StoreLoader;
    use LibraryLoader;

    private $store_id = 0;

    public function __construct($registry)
    {
        parent::__construct($registry);

        $this->store_id = isset($this->request->get['store_id']) ? (int)$this->request->get['store_id'] : 0;
        $this->loadStore($this->store_id);
        $this->loadLibrary($this->store_id);
    }

    public function index()
    {
        header('Content-Type: application/json');
        $protocol = (isset($_SERVER['SERVER_PROTOCOL']) ? $_SERVER['SERVER_PROTOCOL'] : 'HTTP/1.0');

        $type = $this->getQueryParameter('type');
        if (!$type) {
            header("$protocol 400 Bad Request");
            exit;
        }

        $paramSecret = $this->getQueryParameter('secret');
        $secret = $this->config->get('module_connectix_secret');

        $isSpecialRoute = in_array($type, ['confirm', 'cancel']);
        if (!(int)$this->config->get('module_connectix_enabled') || (!$isSpecialRoute && (empty($paramSecret) || $paramSecret !== $secret)) || ($isSpecialRoute && !$this->getQueryParameter('token'))) {
            header("$protocol 400 Bad Request");
            exit;
        }

        $data = $this->getPayload();
        if (!$isSpecialRoute && is_null($data)) {
            header("$protocol 400 Bad Request");
            exit;
        }

        $action = null;
        if (isset($data['action'])) {
            $action = $data['action'];
            unset($data['action']);
        }

        $success = null;
        switch ($type) {
            case 'callback':
                $success = $this->handleCallback($data);
                break;
            case 'inbound':
                $success = $this->handleInbound($data);
                break;
            case 'template':
                $success = $this->handleTemplate($data);
                break;
            case 'tracking':
                if ($action === 'readyForDelivery') {
                    $success = $this->handleReadyForDelivery($data);
                } elseif ($action === 'delivered') {
                    $success = $this->handleDelivered($data);
                } elseif ($action === 'notPickedUp') {
                    $success = $this->handleNotPickedUp($data);
                }
                break;
            case 'confirm':
            case 'cancel':
                $this->handleOrderConfirmation($type, $this->getQueryParameter('token'));
                $success = true;
                break;
            default:
                $success = false;
                break;
        }

        if (true === $success) {
            echo json_encode("OK");
        } elseif (null === $success) {
            header("$protocol 202 Accepted");
            echo json_encode("OK");
        } else {
            header("$protocol 400 Bad Request");
            echo json_encode("Something is wrong..");
        }
        exit;
    }

    /**
     * @return mixed|null
     */
    private function getPayload()
    {
        $payload = file_get_contents('php://input');
        if (empty($payload)) {
            return null;
        }

        return json_decode($payload, true);
    }

    /**
     * @param $name
     * @return string|null
     */
    private function getQueryParameter($name)
    {
        return isset($this->request->get[$name]) ? $this->request->get[$name] : null;
    }

    /**
     * @param $expected
     * @param $data
     * @return bool
     */
    private function validateData($expected, $data)
    {
        $expected = array_flip($expected);
        return !array_diff_key($expected, $data) && !array_diff_key($expected, $data);
    }

    /**
     * @param array $data
     * @return bool
     */
    private function handleCallback(array $data)
    {
        if (!$this->validateData(['id', 'status', 'channel', 'createdAt', 'fallbackId'],  $data) && !$this->validateData(['id', 'status', 'channel', 'createdAt'],  $data)) {
            return false;
        }

        $this->load->model('extension/module/connectix');
        $message = $this->model_extension_module_connectix->getMessage($data['id']);
        if (!$message) {
            return true;
        }

        $this->model_extension_module_connectix->updateMessage($data);

        if (in_array($data['status'], ['delivered', 'seen']) && $message['fallback_id']) {
            $this->model_extension_module_connectix->deleteMessage($message['fallback_id']);
        }

        return true;
    }

    /**
     * @param array $data
     * @return bool
     */
    private function handleInbound(array $data)
    {
        if (!$this->validateData(['id', 'phone', 'text', 'sentAt'],  $data)) {
            return false;
        }

        $this->load->model('extension/module/connectix');
        $text = strtolower($data['text'] ?? '');
        $message = $this->model_extension_module_connectix->getMessage($data['id']);
        if (!$message) {
            return true;
        }

        $insertData = [
            'message_id' => $message['id'],
            'text' => $text,
            'sent_at' => $data['sentAt'] ?? null,
        ];
        $this->model_extension_module_connectix->setMessageInbounds($insertData);

        return true;
    }

    /**
     * @param array $data
     * @return bool
     */
    private function handleTemplate(array $data)
    {
        if (!$this->validateData(['id', 'media', 'layout', 'name', 'channel', 'text', 'button_label', 'button_url', 'status'],  $data)) {
            return false;
        }

        $this->load->model('extension/module/connectix');
        $template = $this->model_extension_module_connectix->getTemplate($data['id']);

        if ($template) {
            $this->model_extension_module_connectix->updateTemplate($data);
        } else {
            $this->model_extension_module_connectix->setTemplate($data);
        }

        return true;
    }

    /**
     * @param array $data
     * @return bool
     */
    private function handleReadyForDelivery(array $data)
    {
        $expected = ['trackingNumber', 'deliveryType'];
        if (isset($data['deliveryType']) && $data['deliveryType'] === 'door') {
            $expected[] = 'courier';
        } elseif (isset($data['deliveryType']) && $data['deliveryType'] === 'office') {
            $expected[] = 'office';
        }

        if (!$this->validateData($expected,  $data)) {
            return false;
        }

        $this->load->model('extension/module/connectix');
        $trackingOrders = $this->model_extension_module_connectix->getTrackingOrders($data['trackingNumber']);
        if (!$trackingOrders) {
            return true;
        }

        $this->load->model('checkout/order');
        $tmp = [];
        foreach ($trackingOrders as $trackingOrder) {
            $order = $this->model_checkout_order->getOrder((int)$trackingOrder['order_id']);
            if ($order) {
                $tmp[] = [
                    'order' => $order,
                    'deliveryData' => $data,
                ];
            }
        }

        if (!empty($tmp)) {
            $library = $this->registry->get('connectix');
            $library->readyForDelivery(['orders' => $tmp]);
        }

        return true;
    }

    /**
     * @param array $data
     * @return bool
     */
    private function handleDelivered(array $data)
    {
        if (empty($data['trackingNumber'])) {
            return false;
        }

        $deliveredStatus = (int)$this->config->get('module_connectix_deliveredStatus');
        if (!$deliveredStatus) {
            return true;
        }

        $this->load->model('extension/module/connectix');
        $trackingOrders = $this->model_extension_module_connectix->getTrackingOrders($data['trackingNumber']);
        if (!$trackingOrders) {
            return true;
        }

        foreach ($trackingOrders as $trackingOrder) {
            $this->model_extension_module_connectix->updateOrder($trackingOrder['order_id'], $deliveredStatus);
        }

        return true;
    }

    /**
     * @param array $data
     * @return bool
     */
    private function handleNotPickedUp(array $data)
    {
        if (!$this->validateData(['trackingNumber', 'workingDaysPassed', 'office'],  $data)) {
            return false;
        }

        $library = $this->registry->get('connectix');
        $rules = array_filter(
            $library->fetchRules($library->getTypeNotePickedUp()),
            function ($rule) use ($data) {
                return (int)$data['workingDaysPassed'] === (int)$rule['days_not_picked_up'];
            }
        );

        if (empty($data['trackingNumber'])) {
            return false;
        }

        if (empty($rules)) {
            return true;
        }

        $this->load->model('extension/module/connectix');
        $trackingOrders = $this->model_extension_module_connectix->getTrackingOrders($data['trackingNumber']);
        if (!$trackingOrders) {
            return false;
        }

        $this->load->model('checkout/order');
        $tmp = [];
        foreach ($trackingOrders as $trackingOrder) {
            $order = $this->model_checkout_order->getOrder((int)$trackingOrder['order_id']);

            if ($order) {
                $tmp[] = [
                    'order' => $order,
                    'deliveryData' => $data,
                ];
            }
        }

        if (!empty($tmp)) {
            $library->notPickedUp(['orders' => $tmp, 'rules' => $rules]);
        }

        return true;
    }

    /**
     * @param string $action
     * @param string $token
     */
    private function handleOrderConfirmation($action, $token)
    {
        $linksValidity = (int)$this->config->get('module_connectix_confirmationUrlsValidity');
        $expiredUrl = $this->config->get('module_connectix_confirmationUrlsExpiredUrl');;

        $enabled = (int)$this->config->get('module_connectix_'.$action.'Order');
        $orderStatus = (int)$this->config->get('module_connectix_'.$action.'edOrderStatus');
        $redirectUrl = $this->config->get('module_connectix_'.$action.'edOrderUrl');

        if (empty($redirectUrl)) {
            $redirectUrl = Context::getContext()->shop->getBaseURL(true);
        }

        if (!$enabled || !$orderStatus) {
            header("Location: $redirectUrl");
            exit;
        }

        $this->load->model('extension/module/connectix');
        $data = $this->model_extension_module_connectix->getOrderByToken($token);
        if ($data) {
            $diff = (time() - strtotime($data['created_at'])) / 3600;
            if (0 < $linksValidity && $linksValidity < $diff) {
                if (!$expiredUrl) {
                    echo json_encode('The link expired');
                }

                header("Location: $expiredUrl");
                exit;
            }

            $this->model_extension_module_connectix->updateOrder($data['order_id'], $orderStatus);
            $this->model_extension_module_connectix->updateOrderToken($token, [
                'used' => 1,
                'status' => $action.'ed',
                'updated_at' => (new DateTime('now'))->format('Y-m-d H:i:s'),
            ]);
        }

        header("Location: $redirectUrl");
        exit;
    }
}
