<?php

namespace ctx;

use Connectix\Client;
use Connectix\Components\Blacklist;
use Connectix\Components\Contact;
use Connectix\Components\Fallback;
use Connectix\Components\Message;
use Connectix\Request\BaseRequest;
use Connectix\Exception;
use DateTime;

require_once __DIR__.'/vendor/autoload.php';

class ConnectixLib
{
    const TYPE_NEW_ORDER = 'new_order';
    const TYPE_CHANGE_ORDER_STATUS = 'change_order_status';
    const TYPE_READY_FOR_DELIVERY = 'ready_for_delivery';
    const TYPE_NOT_PICKED_UP = 'not_picked_up';

    // System type
    const SYSTEM_SANDBOX = 'sandbox';
    const SYSTEM_LIVE = 'live';

    private $registry;
    private $store_id = 0;
    private $client;

    public function __construct($registry, $store_id)
    {
        $this->registry = $registry;
        $this->store_id = $store_id;
    }

    public function __get($key)
    {
        return $this->registry->get($key);
    }

    public function __set($key, $value)
    {
        $this->registry->set($key, $value);
    }

    public function getTypeNewOrder()
    {
        return self::TYPE_NEW_ORDER;
    }

    public function getTypeChangeOrderStatus()
    {
        return self::TYPE_CHANGE_ORDER_STATUS;
    }

    public function getTypeReadyForDelivery()
    {
        return self::TYPE_READY_FOR_DELIVERY;
    }

    public function getTypeNotePickedUp()
    {
        return self::TYPE_NOT_PICKED_UP;
    }

    /**
     * @param $token
     * @param array $data
     * @return mixed
     */
    public function register($token, array $data)
    {
        if (empty($token)) {
            return false;
        }

        return $this->call($token, 'integration/register', $data);
    }

    /**
     * @param $token
     * @param array $data
     * @return mixed
     */
    public function unregister($token, array $data)
    {
        if (empty($token)) {
            return false;
        }

        return $this->call($token, 'integration/unregister', $data);
    }

    /**
     * @param $token
     * @param array $data
     * @return mixed
     */
    public function registerHook($token, array $data)
    {
        if (empty($token)) {
            return false;
        }

        return $this->call($token, 'integration/register/hook', $data);
    }

    /**
     * @param $token
     * @param array $data
     * @return mixed
     */
    public function unregisterHook($token, array $data)
    {
        if (empty($token)) {
            return false;
        }

        return $this->call($token, 'integration/unregister/hook', $data);
    }

    /**
     * @param $token
     * @param $action
     * @param array $data
     * @param array $params
     * @return bool
     */
    private function call($token, $action, $data = [], $params = [])
    {
        $headers = [
            "Content-Type: application/json",
            "Authorization: {$token}",
        ];

        $url = BaseRequest::SANDBOX_ENDPOINT."{$action}";

        $system = isset($data['system']) ? $data['system'] : $this->config->get('module_connectix_system');
        if ($system && $system === self::SYSTEM_LIVE) {
            $url = BaseRequest::LIVE_ENDPOINT."{$action}";
        }
        if (!empty($params)) {
            $url .= '?'.http_build_query($params);
        }

        if (isset($data['system'])) {
            unset($data['system']);
        }

        $ch = curl_init();
        curl_setopt_array($ch, [
            CURLOPT_URL => $url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_FOLLOWLOCATION => false,
            CURLOPT_HTTPHEADER => $headers,
            CURLOPT_TIMEOUT => 3,
            CURLOPT_POST => true,
            CURLOPT_POSTFIELDS => json_encode($data),
        ]);

        $error = false;
        if (!$result = curl_exec($ch)) {
            $error = true;
        }

        if (200 !== curl_getinfo($ch, CURLINFO_RESPONSE_CODE) || curl_errno($ch) !== 0) {
            $error = true;
        }

        curl_close($ch);

        if ($error) {
            return false;
        }

        return json_decode($result, true);
    }

    /**
     * @return Client
     */
    private function getClient()
    {
        if (!$this->client) {
            $this->load->model('setting/setting');
            $settings = $this->model_setting_setting->getSetting('module_connectix', $this->store_id);
            if (!isset($settings['module_connectix_token'])) {
                $settings['module_connectix_token'] = null;
            }

            try {
                $this->client = new Client(
                    $settings['module_connectix_token'],
                    $settings['module_connectix_system'] === self::SYSTEM_SANDBOX
                );
            } catch (Exception\NoTokenException $e) {
            }
        }

        return $this->client;
    }

    /**
     * @return array
     */
    public function syncTemplates()
    {
        try {
            $templates = $this->getClient()->getTemplates()->fetchTemplates();
        } catch (Exception\BadConnectionException $e) {
            return [];
        } catch (Exception\BadRequestException $e) {
            return [];
        } catch (Exception\BadResponseBodyException $e) {
            return [];
        }

        if (!is_array($templates)) {
            return [];
        }

        $this->load->model('extension/module/connectix');

        return $this->model_extension_module_connectix->upsertTemplates($templates);
    }

    /**
     * @param $orderId
     * @param $phone
     * @param array $rule
     * @param array $parameters
     * @param Contact|null $contact
     */
    public function sendMessage($orderId, $phone, array $rule, array $parameters = [], Contact $contact = null)
    {
        $this->load->model('setting/setting');
        $settings = $this->model_setting_setting->getSetting('module_connectix', $this->store_id);

        $enabled = (int)$settings['module_connectix_enabled'];
        $connectixToken = $settings['module_connectix_token'];

        if (empty($phone) || !$enabled || empty($connectixToken)) {
            return;
        }

        $queryData = [];
        $pluginToken = $settings['module_connectix_integrationToken'];
        if (!empty($pluginToken)) {
            $queryData = [
                'integration' => $pluginToken,
            ];
        }

        $client = $this->getClient();
        $sandboxUrl = $settings['module_connectix_sandboxUrl'];
        $isSandbox = $settings['module_connectix_system'] === self::SYSTEM_SANDBOX;

        if ($rule['fallback']) {
            $fallback = new Fallback();
            $fallback->setPhone($phone);
            if ($contact) {
                $fallback->setContact($contact);
            }
            foreach ($rule['templates'] as $tpl) {
                $message = new Message();
                $message->setTemplate($tpl['provider_id'])->setParameters($parameters);

                if ($isSandbox) {
                    $message->setCallbackUrl($sandboxUrl)->setInboundUrl($sandboxUrl);
                }

                $fallback->addMessage($message);
            }

            try {
                $response = $client->getFallback()->send($fallback, $queryData);
            } catch (\Exception $e) {
                // Something went wrong... decide how to handle exceptions...
                return;
            }

            if (!is_array($response)) {
                return;
            }
            foreach ($response['flow'] as $item) {
                $item['fallback_id'] = $response['id'];
                $item['type'] = $rule['type'];
                if ($item['type'] === self::TYPE_READY_FOR_DELIVERY) {
                    $item['delivery_type'] = $rule['delivery_type'];
                }
                $this->writeMessage($orderId, $item);
            }
        } else {
            $request = $client->getMessages();
            foreach ($rule['templates'] as $tpl) {
                $message = new Message();
                $message->setTemplate($tpl['provider_id'])->setPhone($phone)->setParameters($parameters);

                if ($isSandbox) {
                    $message->setCallbackUrl($sandboxUrl)->setInboundUrl($sandboxUrl);
                }

                if ($contact) {
                    $message->setContact($contact);
                }

                try {
                    $response = $request->send($message, $queryData);
                } catch (\Exception $e) {
                    // Something went wrong... decide how to handle exceptions...
                    // Basically same thing as the fallback exception above for the fallback
                    return;
                }
                if (!is_array($response)) {
                    return;
                }

                $response['type'] = $rule['type'];
                if ($response['type'] === self::TYPE_READY_FOR_DELIVERY) {
                    $response['delivery_type'] = $rule['delivery_type'];
                }
                $this->writeMessage($orderId, $response);
            }
        }
    }

    /**
     * @param null $type
     * @param null $id
     * @return mixed
     */
    public function fetchRules($type = null, $id = null)
    {
        $this->load->model('extension/module/connectix');

        return $this->model_extension_module_connectix->fetchRules($type, $id);
    }

    /**
     * @param $token
     * @return mixed
     */
    public function doesOrderTokenExists($token)
    {
        $this->load->model('extension/module/connectix');

        return $this->model_extension_module_connectix->doesOrderTokenExists($token);
    }

    /**
     * @param $orderId
     * @param $token
     */
    public function addOrderToken($orderId, $token)
    {
        $this->load->model('extension/module/connectix');

        $now = (new DateTime('now'))->format('Y-m-d H:i:s');
        $this->model_extension_module_connectix->addOrderToken([
            'order_id' => $orderId,
            'token' => $token,
            'created_at' => $now,
            'updated_at' => $now,
        ]);
    }

    /**
     * @param $orderId
     * @param array $data
     */
    public function writeMessage($orderId, $data)
    {
        $params = [
            'order_id' => $orderId,
            'channel' => $data['channel'],
            'status' => $data['status'],
            'text' => $data['message']['text'],
            'created_at' => $data['createdAt'],
            'provider_id' => $data['id'],
            'type' => $data['type'],
        ];

        if (isset($data['fallback_id'])) {
            $params['fallback_id'] = $data['fallback_id'];
        }

        if ($data['type'] === self::TYPE_READY_FOR_DELIVERY) {
            $params['delivery_type'] = $data['delivery_type'];
        }

        $this->load->model('extension/module/connectix');
        $this->model_extension_module_connectix->writeMessage($params);
    }


    /**
     * @param array $params
     */
    public function readyForDelivery(array $params)
    {
        $rules = $this->fetchRules(self::TYPE_READY_FOR_DELIVERY);
        if (empty($rules)) {
            return;
        }

        $doorRules = array_filter(
            $rules,
            function ($rule) {
                return 'door' === $rule['delivery_type'];
            }
        );
        $officeRules = array_filter(
            $rules,
            function ($rule) {
                return 'office' === $rule['delivery_type'];
            }
        );

        foreach ($params['orders'] as $tmp) {
            $order = $tmp['order'];
            $deliveryData = $tmp['deliveryData'];

            $phone = $order['telephone'];
            if ($this->isPhoneBlacklisted($phone)) {
                continue;
            }

            $parameters = $this->getDefaultParameters($order);
            $parameters['tracking_number'] = $deliveryData['trackingNumber'];

            $contact = new Contact();
            $contact->setFirstName($order['first_name'])
                ->setLastName($order['last_name']);

            if ('office' === $deliveryData['deliveryType']) {
                $parameters['office'] = $deliveryData['office'];
                foreach ($officeRules as $rule) {
                    $this->sendMessage($order['order_id'], $phone, $rule, $parameters, $contact);
                }
            } elseif ('door' === $deliveryData['deliveryType']) {
                $parameters['courier'] = $deliveryData['courier'];
                foreach ($doorRules as $rule) {
                    $this->sendMessage($order['order_id'], $phone, $rule, $parameters, $contact);
                }
            }
        }
    }

    /**
     * @param array $params
     */
    public function notPickedUp(array $params)
    {
        if (!(int)$this->config->get('module_connectix_enabled')) {
            return;
        }

        foreach ($params['orders'] as $tmp) {
            $order = $tmp['order'];
            $deliveryData = $tmp['deliveryData'];

            $phone = $order['telephone'];
            if ($this->isPhoneBlacklisted($phone)) {
                continue;
            }

            $parameters = $this->getDefaultParameters($order);
            $parameters['tracking_number'] = $deliveryData['trackingNumber'];
            $parameters['office'] = $deliveryData['office'];
            $parameters['days_not_picked_up'] = (string)$deliveryData['workingDaysPassed'];

            $contact = new Contact();
            $contact->setFirstName($parameters['first_name'])
                ->setLastName($parameters['last_name']);

            foreach ($params['rules'] as $rule) {
                $this->sendMessage($order['order_id'], $phone, $rule, $parameters, $contact);
            }
        }
    }

    /**
     * @param $phone
     * @return bool
     */
    public function isPhoneBlacklisted($phone)
    {
        $blacklist = new Blacklist();
        $blacklist->setPhone($phone);

        try {
            $isBlacklisted = $this->getClient()->getBlacklist()->isBlacklisted($blacklist);
        } catch (Exception\BadConnectionException $e) {
            return false;
        } catch (Exception\BadRequestException $e) {
            return false;
        } catch (Exception\BadResponseBodyException $e) {
            return false;
        }

        return $isBlacklisted && 'yes' === $isBlacklisted;
    }

    /**
     * @param $trackingNumber
     */
    public function actionObjectOrderCarrierUpdate($trackingNumber)
    {
        if (!(int)$this->config->get('module_connectix_enabled') || !(int)$this->config->get('module_connectix_trackDelivery')) {
            return;
        }

        if (empty($trackingNumber)) {
            return;
        }

        if (is_array($trackingNumber)) {
            $trackingNumber = $trackingNumber[0];
        }

        if (is_object($trackingNumber) && $trackingNumber instanceof \SimpleXMLElement) {
            $trackingNumber = (string)$trackingNumber;
        }
        $data = [
            'trackingNumber' => $trackingNumber,
        ];
        $params = [
            'integration' => $this->config->get('module_connectix_integrationToken'),
        ];

        $this->call($this->config->get('module_connectix_token'), 'trackings', $data, $params);
    }

    public function getDefaultParameters($orderInfo)
    {
        $formatPrice = function ($amount) {
            return $this->currency->format($amount, $this->session->data['currency']);
        };

        $formattedProducts = [];
        $productsCount = 0;
        foreach ($this->cart->getProducts() as $product) {
            $productName = $product['name'];
            if (50 < mb_strlen($productName)) {
                $productName = mb_substr($productName, 0, 47).'...';
            }

            $quantity = (int)$product['quantity'];
            $price = $formatPrice($quantity * $product['price']);
            $productsCount += $quantity;
            $formattedProducts[] = "{$productName} x {$quantity} = {$price}";
        }

        $totals = array();
        $taxes = $this->cart->getTaxes();
        $total = 0;
        $total_data = array(
            'totals' => &$totals,
            'taxes' => &$taxes,
            'total' => &$total,
        );

        $sort_order = array();
        $this->load->model('extension/extension');
        $results = $this->model_extension_extension->getExtensions('total');

        foreach ($results as $key => $value) {
            $sort_order[$key] = $this->config->get('total_'.$value['code'].'_sort_order');
        }

        array_multisort($sort_order, SORT_ASC, $results);

        foreach ($results as $result) {
            if ($this->config->get('total_'.$result['code'].'_status')) {
                $this->load->model('extension/total/'.$result['code']);

                // We have to put the totals in an array so that they pass by reference.
                $this->{'model_extension_total_'.$result['code']}->getTotal($total_data);
            }
        }

        $sort_order = array();
        foreach ($totals as $key => $value) {
            $sort_order[$key] = $value['sort_order'];
        }

        array_multisort($sort_order, SORT_ASC, $totals);

        $totals = array_map(
            function ($total) use ($formatPrice) {
                return "{$total['title']}: {$formatPrice($total['value'])}";
            },
            $totals
        );

        return [
            'order_number' => (string)$orderInfo['order_id'],
            'first_name' => (string)$orderInfo['firstname'],
            'last_name' => (string)$orderInfo['lastname'],
            'total_amount' => $formatPrice($orderInfo['total']),
            'products' => implode("\n\n", $formattedProducts),
            'totals' => implode("\n", $totals),
            'payment_method' => $orderInfo['payment_method'],
            'delivery_type' => $orderInfo['shipping_method'],
            'delivery_address' => $orderInfo['shipping_address_1'].' '.$orderInfo['shipping_city'],
            'products_count' => (string)$productsCount,
        ];
    }
}
